//
//$Id: LogServer.java 9 2005-05-25 20:06:23Z olef $
//

package performancemonster.loghandler;

import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.List;

/**
 * A daemon thread the processes connections from a
 * <code>org.apache.log4j.net.SocketAppender.html</code>.
 *
 * @author <a href="mailto:oliver@puppycrawl.com">Oliver Burn</a>
 */
public class LogServer extends Thread {
    /**
     * Helper that actually processes a client connection. It receives events
     * and adds them to the supplied model.
     *
     * @author <a href="mailto:oliver@puppycrawl.com">Oliver Burn</a>
     */
    private class Slurper implements Runnable {
        /** socket connection to read events from **/
        private final Socket mClient;

        /**
         * Creates a new <code>Slurper</code> instance.
         *
         * @param aClient socket to receive events from
         */
        Slurper(Socket aClient) {
            mClient = aClient;
        }

        /** loops getting the events **/
        public void run() {
            System.out.println("Starting to get data");
            try {
                final ObjectInputStream ois =
                    new ObjectInputStream(mClient.getInputStream());
                while (true) {
                    final LoggingEventWrapper eventWrapper = (LoggingEventWrapper) ois.readObject();
                    notifyListeners(eventWrapper);
                }
            } catch (EOFException e) {
                //LOG.info("Reached EOF, closing connection");
            } catch (SocketException e) {
                //LOG.info("Caught SocketException, closing connection");
            } catch (IOException e) {
                //LOG.warn("Got IOException, closing connection", e);
            } catch (ClassNotFoundException e) {
                //LOG.warn("Got ClassNotFoundException, closing connection", e);
            }

            try {
                mClient.close();
            } catch (IOException e) {
                //LOG.warn("Error closing connection", e);
            }
        }
    }

    /** server for listening for connections **/
    private final ServerSocket mSvrSock;

    private List<LogListener> logListeners;
    
    /**
     * Creates a new <code>LoggingReceiver</code> instance.
     *
     * @param aModel model to place put received into
     * @param aPort port to listen on
     * @throws IOException if an error occurs
     */
    public LogServer(int aPort) throws IOException {
        setDaemon(true);
        mSvrSock = new ServerSocket(aPort);
        logListeners = new ArrayList<LogListener>();
    }
    
    public void addLogListener(LogListener listener) {
        logListeners.add(listener);
    }

    /** Listens for client connections **/
    public void run() {
        //LOG.info("Thread started");
        try {
            while (true) {
                System.out.println("Waiting for a connection");
                final Socket client = mSvrSock.accept();
                System.out.println("Got a connection from " +
                          client.getInetAddress().getHostName());
                final Thread t = new Thread(new Slurper(client));
                t.setDaemon(true);
                t.start();
            }
        } catch (IOException e) {
            //LOG.error("Error in accepting connections, stopping.", e);
        }
    }
    
    private void notifyListeners(LoggingEventWrapper eventWrapper) {
        for (LogListener listener : logListeners) {
            listener.receiveLogEvent(eventWrapper);
        }
    }
}
