//
//$Id: MonsterAppender.java 9 2005-05-25 20:06:23Z olef $
//

package performancemonster.loghandler;

import java.net.InetAddress;
import java.net.Socket;
import java.io.IOException;
import java.io.ObjectOutputStream;

import org.apache.log4j.helpers.LogLog;
import org.apache.log4j.spi.LoggingEvent;
import org.apache.log4j.AppenderSkeleton;


public class MonsterAppender extends AppenderSkeleton {

  /**
     The default port number of remote logging server (4560).
  */
  static final int DEFAULT_PORT                 = 4560;

  /**
     The default reconnection delay (30000 milliseconds or 30 seconds).
  */
  static final int DEFAULT_RECONNECTION_DELAY   = 30000;

  /**
     We remember host name as String in addition to the resolved
     InetAddress so that it can be returned via getOption().
  */
  String remoteHost;

  InetAddress address;
  int port = DEFAULT_PORT;
  ObjectOutputStream oos;
  int reconnectionDelay = DEFAULT_RECONNECTION_DELAY;

  private Connector connector;

  int counter = 0;


  // reset the ObjectOutputStream every 70 calls
  //private static final int RESET_FREQUENCY = 70;
  private static final int RESET_FREQUENCY = 1;

  public MonsterAppender() {
  }

  /**
     Connect to the specified <b>RemoteHost</b> and <b>Port</b>.
  */
  public void activateOptions() {
    connect(address, port);
  }

  /**
   * Close this appender.  
   *
   * <p>This will mark the appender as closed and call then {@link
   * #cleanUp} method.
   * */
  synchronized public void close() {
        if (closed)
            return;

        this.closed = true;
        cleanUp();
    }

  /**
   * Drop the connection to the remote host and release the underlying
   * connector thread if it has been created 
   * */
  public void cleanUp() {
        if (oos != null) {
            try {
                oos.close();
            } catch (IOException e) {
                LogLog.error("Could not close oos.", e);
            }
            oos = null;
        }
        if (connector != null) {
            // LogLog.debug("Interrupting the connector.");
            connector.interrupted = true;
            connector = null; // allow gc
        }
    }

  void connect(InetAddress address, int port) {
        if (this.address == null)
            return;
        try {
            // First, close the previous connection if any.
            cleanUp();
            oos = new ObjectOutputStream(new Socket(address, port)
                    .getOutputStream());
        } catch (IOException e) {

            String msg = "Could not connect to remote log4j server at ["
                    + address.getHostName() + "].";
            if (reconnectionDelay > 0) {
                msg += " We will try again later.";
                fireConnector(); // fire the connector thread
            }
            LogLog.error(msg, e);
        }
    }


  public void append(LoggingEvent event) {
        if (event == null)
            return;
        
        LoggingEventWrapper eventWrapper = new LoggingEventWrapper(event);

        if (address == null) {
            errorHandler
                    .error("No remote host is set for SocketAppender named \""
                            + this.name + "\".");
            return;
        }

        if (oos != null) {
            try {
                oos.writeObject(eventWrapper);
                // LogLog.debug("=========Flushing.");
                oos.flush();
                if (++counter >= RESET_FREQUENCY) {
                    counter = 0;
                    // Failing to reset the object output stream every now and
                    // then creates a serious memory leak.
                    // System.err.println("Doing oos.reset()");
                    oos.reset();
                }
            } catch (IOException e) {
                oos = null;
                LogLog.warn("Detected problem with connection: " + e);
                if (reconnectionDelay > 0) {
                    fireConnector();
                }
            }
        }
    }

  void fireConnector() {
    if(connector == null) {
      LogLog.debug("Starting a new connector thread.");
      connector = new Connector();
      connector.setDaemon(true);
      connector.setPriority(Thread.MIN_PRIORITY);
      connector.start();
    }
  }

  static
  InetAddress getAddressByName(String host) {
    try {
      return InetAddress.getByName(host);
    } catch(Exception e) {
      LogLog.error("Could not find address of ["+host+"].", e);
      return null;
    }
  }

  /**
   * The SocketAppender does not use a layout. Hence, this method
   * returns <code>false</code>.  
   * */
  public boolean requiresLayout() {
    return false;
  }

  /**
   * The <b>RemoteHost</b> option takes a string value which should be
   * the host name of the server where a {@link SocketNode} is
   * running.
   * */
  public void setRemoteHost(String host) {
    address = getAddressByName(host);
    remoteHost = host;
  }

  /**
     Returns value of the <b>RemoteHost</b> option.
   */
  public String getRemoteHost() {
    return remoteHost;
  }

  /**
     The <b>Port</b> option takes a positive integer representing
     the port where the server is waiting for connections.
   */
  public void setPort(int port) {
    this.port = port;
  }

  /**
     Returns value of the <b>Port</b> option.
   */
  public int getPort() {
    return port;
  }

  /**
     The <b>ReconnectionDelay</b> option takes a positive integer
     representing the number of milliseconds to wait between each
     failed connection attempt to the server. The default value of
     this option is 30000 which corresponds to 30 seconds.

     <p>Setting this option to zero turns off reconnection
     capability.
   */
  public void setReconnectionDelay(int delay) {
    this.reconnectionDelay = delay;
  }

  /**
     Returns value of the <b>ReconnectionDelay</b> option.
   */
  public int getReconnectionDelay() {
    return reconnectionDelay;
  }

  /**
     The Connector will reconnect when the server becomes available
     again.  It does this by attempting to open a new connection every
     <code>reconnectionDelay</code> milliseconds.

     <p>It stops trying whenever a connection is established. It will
     restart to try reconnect to the server when previpously open
     connection is droppped.

     @author  Ceki G&uuml;lc&uuml;
     @since 0.8.4
  */
  class Connector extends Thread {

    boolean interrupted = false;

    public
    void run() {
      Socket socket;
      while(!interrupted) {
    try {
      sleep(reconnectionDelay);
      LogLog.debug("Attempting connection to "+address.getHostName());
      socket = new Socket(address, port);
      synchronized(this) {
        oos = new ObjectOutputStream(socket.getOutputStream());
        connector = null;
        LogLog.debug("Connection established. Exiting connector thread.");
        break;
      }
    } catch(InterruptedException e) {
      LogLog.debug("Connector interrupted. Leaving loop.");
      return;
    } catch(java.net.ConnectException e) {
      LogLog.debug("Remote host "+address.getHostName()
               +" refused connection.");
    } catch(IOException e) {
      LogLog.debug("Could not connect to " + address.getHostName()+
               ". Exception is " + e);
    }
      }
      //LogLog.debug("Exiting Connector.run() method.");
    }

    /**
       public
       void finalize() {
       LogLog.debug("Connector finalize() has been called.");
       }
    */
  }

}
