//
//$Id: GraphArea.java 9 2005-05-25 20:06:23Z olef $
//

package performancemonster.presentation;

import java.awt.*;
import java.awt.event.MouseEvent;

import javax.swing.JLabel;
import javax.swing.event.MouseInputAdapter;

import java.util.ArrayList;
import java.util.List;

public class GraphArea extends JLabel {
    private List<GraphElement> graphElements = new ArrayList<GraphElement>();
    private List<GraphElement> shownElements;
    private List<DescriptionListener> descriptionListeners = new ArrayList<DescriptionListener>();
    private List<HoverListener> hoverListeners = new ArrayList<HoverListener>();
    private long beginning;
    private long end;
    private boolean zoomMode;
    private Rectangle currentRect = null;
    Rectangle rectToDraw = null;
    Rectangle previousRectDrawn = new Rectangle();

    public GraphArea() {
        setOpaque(true);
        setMinimumSize(new Dimension(10,10));
        setPreferredSize(new Dimension(100, 100));

        MyListener myListener = new MyListener();
        addMouseListener(myListener);
        addMouseMotionListener(myListener);
    }

    public void setGraphElements(List<GraphElement> graphElements) {
        this.graphElements = graphElements;
        eraseCaches();
    }
    
    public void setBeginning(long beginning) {
        this.beginning = beginning;
        eraseCaches();
    }
    
    public void setEnd(long end) {
        this.end = end;
        eraseCaches();
    }
    
    public long getBeginning() {
        return beginning;
    }
    
    public long getEnd() {
        return end;
    }
    
    public void setZoomMode(boolean zoomMode) {
        this.zoomMode = zoomMode;
        repaint();
    }
    
    public boolean getZoomMode() {
        return zoomMode;
    }
    
    public void addDescriptionListener(DescriptionListener listener) {
        descriptionListeners.add(listener);
    }
    
    public void addHoverListener(HoverListener listener) {
        hoverListeners.add(listener);
    }
    
    /**
     * Sets the beginning and end points so that there is just enough space
     * for all graph elements.
     */
    public void align() {
        long newBeginning = Long.MAX_VALUE;
        long newEnd = Long.MIN_VALUE;
        for (GraphElement element : graphElements) {
            newBeginning = Math.min(newBeginning, element.getBeginning());
            newEnd = Math.max(newEnd, element.getEnd());
        }
        setBeginning(newBeginning);
        setEnd(newEnd);
        repaint();
    }
    
    protected void paintComponent(Graphics g) {
        super.paintComponent(g); //paints the background and image

        // Paint graph boxes
        int currentPosition = 0;
        for (GraphElement element : getShownElements()) {
            paintElement(g, element, currentPosition++);
        }
        
        // Paint zoom box
        if (zoomMode && currentRect != null) {
            //Draw a rectangle on top of the image.
            g.setColor(Color.BLACK);
            g.setXORMode(Color.white); //Color of line varies
                                       //depending on image colors
            g.fillRect(rectToDraw.x, 0, 
                       rectToDraw.width, getHeight());
        }
    }

    private GraphElement findElementAt(int x, int y) {
        int height = getHeight();
        for (GraphElement element : getShownElements()) {
            if (x >= convertGraphXToScreen(element.getBeginning())
                    && x <= convertGraphXToScreen(element.getEnd())
                    && y >= height - getElementHeight() && y <= height) {
                return element;
            }
            height -= getElementHeight();
        }
        return null;
    }
    
    private class MyListener extends MouseInputAdapter {
        public void mousePressed(MouseEvent e) {
            int x = e.getX();
            int y = e.getY();
            if (zoomMode) {
                currentRect = new Rectangle(x, y, 0, 0);
                updateDrawableRect(getWidth(), getHeight());
                repaint();
            } else {
                GraphElement clickedElement = findElementAt(x, y);
                if (clickedElement != null) {
                    for (DescriptionListener listener : descriptionListeners) {
                        listener.showDescription(clickedElement);
                    }
                }
                notifyHoverListeners(x, y);
            }
        }

        public void mouseDragged(MouseEvent e) {
            int x = e.getX();
            int y = e.getY();
            if (zoomMode) {
                updateSize(e);
            } else {
                notifyHoverListeners(x, y);
            }
        }

        public void mouseReleased(MouseEvent e) {
            int x = e.getX();
            int y = e.getY();
            if (zoomMode) {
                updateSize(e);
                setBeginning(convertScreenXToGraph(currentRect.x));
                setEnd(convertScreenXToGraph(currentRect.x + currentRect.width));
                currentRect = null;
                repaint();
            } else {
                notifyHoverListeners(x, y);
            }
        }
        
        public void mouseMoved(MouseEvent e) {
            int x = e.getX();
            int y = e.getY();
            if (zoomMode) {
                // Nothing to do
            } else {
                notifyHoverListeners(x, y);
            }
        }

        /* 
         * Update the size of the current rectangle
         * and call repaint.  Because currentRect
         * always has the same origin, translate it
         * if the width or height is negative.
         * 
         * For efficiency (though
         * that isn't an issue for this program),
         * specify the painting region using arguments
         * to the repaint() call.
         * 
         */
        void updateSize(MouseEvent e) {
            int x = e.getX();
            int y = e.getY();
            currentRect.setSize(x - currentRect.x,
                                y - currentRect.y);
            updateDrawableRect(getWidth(), getHeight());
            repaint();
        }
    }
    
    private void notifyHoverListeners(int x, int y) {
        GraphElement element = null;
        if (getShownElements().size() > 0) {
            element = getShownElements().get(convertScreenYToGraph(y));
        }
        long graphX = convertScreenXToGraph(x);
        for (HoverListener listener : hoverListeners) {
            listener.mouseHovered(graphX, element);
        }
    }
    
    private void eraseCaches() {
        shownElements = null;
    }
    
    private List<GraphElement> getShownElements() {
        if (shownElements == null) {
            extractShownElements();
        }
        return shownElements;
    }

    private void paintElement(Graphics g, GraphElement element, int pos) {
        int x1 = getElementLeft(element);
        int x2 = getElementRight(element);
        int y1 = getElementTop(pos);
        int y2 = getElementBottom(pos);
        
        g.setColor(element.getColor());
        g.fillRect(x1, y1, x2 - x1, y2 - y1);
    }
    
    private void updateDrawableRect(int compWidth, int compHeight) {
        int x = currentRect.x;
        int y = currentRect.y;
        int width = currentRect.width;
        int height = currentRect.height;

        //Make the width and height positive, if necessary.
        if (width < 0) {
            width = 0 - width;
            x = x - width + 1; 
            if (x < 0) {
                width += x; 
                x = 0;
            }
        }
        if (height < 0) {
            height = 0 - height;
            y = y - height + 1; 
            if (y < 0) {
                height += y; 
                y = 0;
            }
        }

        //The rectangle shouldn't extend past the drawing area.
        if ((x + width) > compWidth) {
            width = compWidth - x;
        }
        if ((y + height) > compHeight) {
            height = compHeight - y;
        }
      
        //Update rectToDraw after saving old value.
        if (rectToDraw != null) {
            previousRectDrawn.setBounds(
                        rectToDraw.x, rectToDraw.y, 
                        rectToDraw.width, rectToDraw.height);
            rectToDraw.setBounds(x, y, width, height);
        } else {
            rectToDraw = new Rectangle(x, y, width, height);
        }
    }

    private void extractShownElements() {
        shownElements = new ArrayList<GraphElement>();
        for (GraphElement element : graphElements) {
            if (isElementVisible(element)) {
                shownElements.add(element);
            }
        }
    }
    
    private boolean isElementVisible(GraphElement element) {
        return element.getBeginning() <= getEnd()
                && element.getEnd() >= getBeginning();
    }
    
    private int getElementLeft(GraphElement element) {
        return convertGraphXToScreen(element.getBeginning());
    }

    private int getElementRight(GraphElement element) {
        return convertGraphXToScreen(element.getEnd());
    }

    private int getElementTop(int pos) {
        return (int) (getHeight() - ((pos + 1) * getElementHeight()));
    }

    private int getElementBottom(int pos) {
        return (int) (getElementTop(pos) + getElementHeight());
    }
    
    private double getElementHeight() {
        if (shownElements.size() > 0) {
            return (double) getHeight() / (double) shownElements.size();
        } else {
            // Just some dummy value...
            return 1;
        }
    }
    
    private int convertGraphXToScreen(long x) {
        long xRange = getEnd() - getBeginning();
        return (int) (((x - getBeginning()) * getWidth()) / xRange);
    }
    
    private long convertScreenXToGraph(int x) {
        long xRange = getEnd() - getBeginning();
        return (xRange * x) / getWidth() + getBeginning();
    }
    
    private int convertScreenYToGraph(int y) {
        int maxValue = getShownElements().size() - 1;
        return Math.min(maxValue, maxValue
                - (int) ((double) y / getElementHeight()));
    }
}
